#!/usr/bin/env python
#
#    CassaFS
#    Copyright (C) 2012  Paul Dwerryhouse <paul@dwerryhouse.com.au>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from pycassa.pool import ConnectionPool
from pycassa.columnfamily import ColumnFamily
from pycassa.cassandra.ttypes import NotFoundException
from pycassa.types import *
from time import time
from collections import OrderedDict
import grp, pwd
import uuid, sys

import os, stat, errno, os.path
# pull in some spaghetti to make this stuff work without fuse-py being installed
try:
    import _find_fuse_parts
except ImportError:
    pass
import fuse
from fuse import Fuse

BLKSIZE = 8192

KEYSPACE="CassaFS"
SERVER="127.0.0.1:9160"

POOL = ConnectionPool(KEYSPACE, server_list=['127.0.0.1:9160'])

DIRECTORY = ColumnFamily(POOL, 'Directory')
INODE = ColumnFamily(POOL, 'Inode')
BLOCK = ColumnFamily(POOL, 'Block')

INODE.column_validators['nlink'] = IntegerType()
INODE.column_validators['mode'] = IntegerType()
INODE.column_validators['rdev'] = IntegerType()
INODE.column_validators['uid'] = IntegerType()
INODE.column_validators['gid'] = IntegerType()
INODE.column_validators['size'] = IntegerType()
INODE.column_validators['blksize'] = IntegerType()
INODE.column_validators['blocks'] = IntegerType()
INODE.column_validators['mtime'] = IntegerType()
INODE.column_validators['ctime'] = IntegerType()
INODE.column_validators['atime'] = IntegerType()

INODE.key_validation_class = LexicalUUIDType()
BLOCK.key_validation_class = LexicalUUIDType()
DIRECTORY.key_validation_class = LexicalUUIDType()

DIRECTORY.default_validation_class = LexicalUUIDType()

BLOCK.column_name_class = IntegerType()

rootkey = uuid.UUID('{00000000-0000-0000-0000-000000000000}')

if not hasattr(fuse, '__version__'):
    raise RuntimeError, \
        "your fuse-py doesn't know of fuse.__version__, probably it's too old."

fuse.fuse_python_api = (0, 2)


def replace(s,buf,p):
    ns = ""
    n = len(s)
    bn = len(buf)
    if n <= p:
        ns = s + ('\0' * (p-n)) + buf
    else:
        ns = s[0:p] + buf + s[p+bn:]
    return ns

def writeblock(key, buf, blkid, pos):
    try:
        block = BLOCK.get(key, column_start=blkid, column_finish=blkid)
    except NotFoundException:
        block = OrderedDict([(blkid, b'')])
    
    newblock = replace(block[blkid],buf,pos)
    BLOCK.insert(key,{ blkid: newblock })

    return len(buf)

def user_in_group(uid,gid):
    
    try:
        if pwd.getpwuid(uid).pw_gid == gid:
            return True
    except KeyError:
        pass

    for user in grp.getgrgid(gid).gr_mem:
        if uid == pwd.getpwnam(user).pw_uid:
            return True

    return False
    
def get_path_components(path):
    paths = [ path ]
    while path != '/':
        path = os.path.dirname(path)
        paths.insert(0,path)
    try:
        end = paths.pop()
    except IndexError:
        end = ''

    try:
        notend = paths.pop()
    except IndexError:
        notend = ''
    
    return (end, notend, paths)
    
class MyStat(fuse.Stat):
    def __init__(self):
        self.st_mode = 0
        self.st_ino = 0
        self.st_dev = 0
        self.st_nlink = 0
        self.st_uid = 0
        self.st_gid = 0
        self.st_size = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0

def get_key(path):
    key = None
    for p in path.split('/'):
        if p == '':
            key = rootkey
        else:
            try:
                key = DIRECTORY.get(key, columns=[p])[p]
            except KeyError:
                raise NotFoundException
    return key
    
class CassaFS(Fuse):

    def getattr(self, path):

        st = MyStat()
        try:
            key = get_key(path)
            inode = INODE.get(key)
            if inode.get('mode') != None:
                st.st_mode = inode.get('mode')
            if inode.get('nlink') != None:
                st.st_nlink = inode.get('nlink')
            if inode.get('rdev') != None:
                st.st_rdev = inode.get('rdev')
            if inode.get('uid') != None:
                st.st_uid = inode.get('uid')
            if inode.get('gid') != None:
                st.st_gid = inode.get('gid')
            if inode.get('atime') != None:
                st.st_atime = inode.get('atime')
            if inode.get('mtime') != None:
                st.st_mtime = inode.get('mtime')
            if inode.get('ctime') != None:
                st.st_ctime = inode.get('ctime')
            if inode.get('blocks') != None:
                st.st_blocks = inode.get('blocks')
            if inode.get('size') != None:
                st.st_size = inode.get('size')
        except NotFoundException:
            return -errno.ENOENT

        return st

    def check_access(self, path, mode):

        try:
            key = get_key(path)
        except NotFoundException:
            return -errno.EEXIST

        if mode == os.F_OK:
            return 0

        inode = INODE.get(key)
        fmode = inode['mode']
        context = self.GetContext()
        uid = context['uid']
        gid = context['gid']


        if (inode['uid'] == uid):
            if (mode & os.X_OK) and (fmode & stat.S_IXUSR):
                return 0

            if (mode & os.W_OK) and (fmode & stat.S_IWUSR):
                return 0

            if (mode & os.R_OK) and (fmode & stat.S_IRUSR):
                return 0

        if (inode['gid'] == gid):
            if (mode & os.X_OK) and (fmode & stat.S_IXGRP):
                return 0

            if (mode & os.W_OK) and (fmode & stat.S_IWGRP):
                return 0

            if (mode & os.R_OK) and (fmode & stat.S_IRGRP):
                return 0

        if (mode & os.X_OK) and (fmode & stat.S_IXOTH):
            return 0

        if (mode & os.W_OK) and (fmode & stat.S_IWOTH):
            return 0

        if (mode & os.R_OK) and (fmode & stat.S_IROTH):
            return 0
        
        return -errno.EACCES

    
    def access(self, path, mode):
        print "ACCESS {0} {1}".format(path,mode)

        (first, second, paths) = get_path_components(path)

        # Check that we can access leading paths
        for p in paths:
            err = self.check_access(p, os.X_OK)
            if err < 0:
                return err

        if second != '':
            err = self.check_access(second, os.X_OK)
            if err < 0:
                return err

        err = self.check_access(first, mode)
        if err < 0:
            return err

        return 0


    def readdir(self, path, offset):
        # XXX currently ignoring the offset

        # Not entirely sure if sys.maxint is the largest number of
        # columns per row
        key = get_key(path)
        dirs = DIRECTORY.get(key,column_count=sys.maxint)
        for r in dirs:
            yield fuse.Direntry(r)

    def chown(self, path, uid, gid):
        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)

            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err

        key = get_key(path)
        inode = INODE.get(key)

        if gid == -1:
            if context['uid'] != 0:
                return -errno.EPERM

            inode = {
                'uid': uid,
            }

        elif uid == -1:
            if context['uid'] != 0:
                if context['uid'] != inode['uid']:
                    return -errno.EPERM

                if context['gid'] != gid and not user_in_group(context['uid'],gid):
                    return -errno.EPERM

            inode = {
                'gid': gid,
            }

        else:
            if context['uid'] != 0:
                return -errno.EPERM

            inode = {
                'uid': uid,
                'gid': gid,
            }
        INODE.insert(key,inode)
        

    def chmod(self, path, mode):
        print "###         CHMOD {0} {1}\n".format(path,mode)

        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK)
                if err < 0:
                    return err

        key = get_key(path)
        inode = INODE.get(key)

        if context['uid'] != inode['uid'] and context['uid'] != 0:
            return -errno.EPERM
    
        oldmode = inode.get('mode')
        mode = (oldmode & 0770000) | ( mode & 07777)
        inode = {
            'mode': mode
        }
        INODE.insert(key,inode)
        
    def mknod(self, path, mode, dev):
        key = uuid.uuid4()
        now = int(time())

        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = get_key(dirname)

        inode = {
            'uid': 0,
            'gid': 0,
            'rdev': dev,
            'mode': mode,
            'size': 0,
            'blocks': 0,
            'blksize': 0,
            'nlink': 1,
            'atime': now,
            'mtime': now,
            'ctime': now,
        }
        INODE.insert(key,inode)
        DIRECTORY.insert(parentkey, {basename : key})

    def rmdir(self, path):
        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

        key = get_key(path)
        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = get_key(dirname)
        dir = DIRECTORY.get(key)
        if len(dir) > 2:
            return -errno.ENOTEMPTY
        
        DIRECTORY.remove(parentkey, { basename })
        DIRECTORY.remove(key)
        INODE.remove(key)
        
    def mkdir(self, path, mode):
        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

        key = uuid.uuid4()
        now = int(time())
        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = get_key(dirname)
        newdir = { basename: key }

        thisdir = { '.': key }
        parentdir = { '..': parentkey }

        inode = {
            'mode': mode | stat.S_IFDIR,
            'uid': context['uid'],
            'gid': context['gid'],
            'rdev': 0,
            'size': 0,
            'blocks': 0,
            'blksize': 0,
            'nlink': 2,
            'atime': now,
            'mtime': now,
            'ctime': now,
        }
        INODE.insert(key,inode)
        DIRECTORY.insert(key,thisdir)
        DIRECTORY.insert(key,parentdir)
        DIRECTORY.insert(parentkey, newdir)
        
        return 0

    def rename(self, path1, path2):
        key1 = get_key(path1)
        inode1 = INODE.get(key1)

        file1 = os.path.basename(path1)
        dir1 = os.path.dirname(path1)
        err = self.access(dir1, os.W_OK)
        if err < 0:
            return err

        file2 = os.path.basename(path2)
        dir2 = os.path.dirname(path2)
        err = self.access(dir2, os.W_OK)
        if err < 0:
            return err

        dirkey1 = get_key(dir1)
        direntry = DIRECTORY.get(dirkey1, columns=[file1])[file1]
        DIRECTORY.remove(dirkey1, { file1 });
        dirkey2 = get_key(dir2)
        newdirentry = { file2: key1 }
        DIRECTORY.insert(dirkey2, newdirentry)
        
    def unlink(self, path):

        (first, second, paths) = get_path_components(path)
        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(path)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

        key = get_key(path)
        inode = INODE.get(key)
            
        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = get_key(dirname)
        DIRECTORY.remove(parentkey, { basename });
        if inode['nlink'] <= 1:
            INODE.remove(key)
            BLOCK.remove(key)
        else:
            inode['nlink']-=1
            INODE.insert(key,inode)
        
    def readlink(self, path):

        key = get_key(path)
        inode = INODE.get(key)
        return inode['symlink']

    def symlink(self, targetpath, linkpath):
        context = self.GetContext()

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(linkpath)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

        key = uuid.uuid4()
        now = int(time())
        basename = os.path.basename(linkpath)
        dirname = os.path.dirname(linkpath)
        parentkey = get_key(dirname)
        inode = {
            'mode': 777 | stat.S_IFLNK,
            'uid': context['uid'],
            'gid': context['gid'],
            'nlink': 1,
            'rdev': 0,
            'size': 0,
            'blocks': 0,
            'blksize': 0,
            'atime': now,
            'mtime': now,
            'ctime': now,
            'symlink': targetpath
        }
        INODE.insert(key,inode)
        DIRECTORY.insert(parentkey, {basename : key})

    def link(self, targetpath, linkpath):
        context = self.GetContext()
        key = get_key(targetpath)
        inode = INODE.get(key)

        if context['uid'] != 0:
            (first, second, paths) = get_path_components(targetpath)
    
            # Check that we can access leading paths
            for p in paths:
                err = self.check_access(p, os.X_OK)
                if err < 0:
                    return err
    
            if second != '':
                err = self.check_access(second, os.X_OK | os.W_OK)
                if err < 0:
                    return err

            if inode['uid'] != context['uid']:
                return -errno.EACCESS

        basename = os.path.basename(linkpath)
        dirname = os.path.dirname(linkpath)
        parentkey = get_key(dirname)
        inode['nlink'] += 1
        INODE.insert(key,inode)
        DIRECTORY.insert(parentkey, {basename : key})

    def truncate(self, path, size):

        key = get_key(path)
        inode = INODE.get(key)
        total = inode['size']
        blkid = int(size / BLKSIZE)
        endblkid = int(total / BLKSIZE)
        BLOCK.remove(key, range(blkid+1, endblkid+1))
        offset = size % BLKSIZE
        try:
            block = BLOCK.get(key,column_start=blkid,column_finish=blkid)
        except NotFoundException:
            block = OrderedDict([(blkid, b'')])
        newblock = block[blkid][0:offset]
        BLOCK.insert(key, { blkid : newblock })
        INODE.insert(key, { 'size': size })

    def utimens(self, path, atime, mtime):

        # XXX check to see if noatime or relatime alters anything
        key = get_key(path)
        inode = {}
        inode['atime'] = atime.tv_sec
        inode['mtime'] = mtime.tv_sec
        inode = INODE.insert(key,inode)

    def create(self, path, flags, mode):


        (first, second, paths) = get_path_components(path)

        # Check that we can access leading paths
        for p in paths:
            err = self.check_access(p, os.X_OK)
            if err < 0:
                return err

        if second != '':
            err = self.check_access(second, os.X_OK | os.W_OK)
            if err < 0:
                return err

        err = self.check_access(first, mode)
        if err < 0 and err != -errno.EEXIST:
            return err

        context = self.GetContext()
        key = uuid.uuid4()
        now = int(time())
        basename = os.path.basename(path)
        dirname = os.path.dirname(path)
        parentkey = get_key(dirname)
        inode = {
            'mode': mode | stat.S_IFREG,
            'uid': context['uid'],
            'gid': context['gid'],
            'nlink': 1,
            'rdev': 0,
            'size': 0,
            'blocks': 0,
            'blksize': 0,
            'atime': now,
            'mtime': now,
            'ctime': now,
        }
        INODE.insert(key,inode)
        DIRECTORY.insert(parentkey, {basename : key})
        

    def open(self, path, flags):

        # XXX maybe should look to see if O_TRUNC/atomic_o_trunc causes
        # any concern here

        if flags & 0x03 == os.O_RDONLY:
            return self.access(path, os.R_OK)

        elif flags & 0x03 == os.O_WRONLY:
            return self.access(path, os.W_OK)

        elif flags & 0x03 == os.O_RDWR:
            return self.access(path, os.R_OK | os.W_OK)


    def write(self, path, buf, offset):

        key = get_key(path)
        buflen = len(buf)
        n = buflen
        start = int(offset / BLKSIZE)
        finish = int((offset + n - 1) / BLKSIZE)
        inode = INODE.get(key)
        size = inode['size']

        o = offset % BLKSIZE
        p = 0

        print "WRITE {0} bytes at offset {1}".format(buflen,offset)
        for i in range(start,finish+1):
            if (n <= BLKSIZE - o):
                writeblock(key, buf[p:], start, o)
                break
            l = BLKSIZE - o
            writeblock(key, buf[p:p+l], i, o)
            p += l
            n -= l
            o = 0
            
        if (offset + buflen) > size:
            size = offset + buflen
            inode = { 'size': offset + buflen }

        inode['blocks'] = (offset % BLKSIZE) + 1
        INODE.insert(key,inode)
        
        return n

    def read(self, path, size, offset):

        key = get_key(path)
        blkid = 0
        start = int(offset / BLKSIZE)
        finish = int((offset + size - 1) / BLKSIZE)
        try:
            block = BLOCK.get(key, column_start=start, column_finish=finish)
        except NotFoundException:
            return ''
        
        buf = ''
        o = offset % BLKSIZE
        for i in range(start,finish+1):
            if (size <= BLKSIZE - o):
                buf += block[i][o:o+size]
                break
            buf += block[i][o:]
            size -= len(block[i][o:])
            o = 0
            
        return buf

    # XXX this is all clearly bogus
    def statfs(self):

        fs = fuse.StatVfs()
        fs.f_namemax = 255
        fs.f_bsize = BLKSIZE
        fs.f_frsize = BLKSIZE
        fs.f_blocks = 0
        fs.f_bfree = 0
        fs.f_bavail = 0

        return fs

def main():
    usage="""
Userspace CassaFS example

""" + Fuse.fusage
    server = CassaFS(version="%prog " + fuse.__version__,
                     usage=usage,
                     dash_s_do='setsingle')

    server.parse(errex=1)
    server.main()

if __name__ == '__main__':
    main()
